// Copyright Erkki Seppl 2001-2002
// $Id: mediaqueue.cc,v 1.18 2002/05/27 20:53:45 flux Exp $
// $Author: flux $ 
// $Revision: 1.18 $

static char* gs_id __attribute__ ((unused)) ="$Id: mediaqueue.cc,v 1.18 2002/05/27 20:53:45 flux Exp $"; /// ID for CVS/RCS

#include <iostream>
#include <fstream>
#include <vector>
#include <iterator>
#include <functional>

#include <time.h>
#include <string.h>

#include "mediaqueue.hh"
#include "show.hh"
#include "occurrence.hh"
#include "device.hh"

typedef std::multimap<time_t, Show*> TimeShowMMap;

MediaQueue::MediaQueue(DeviceList& a_devices) :
  m_devices(a_devices)
{
  // nothing
}

MediaQueue::~MediaQueue()
{
  for (ShowList::iterator it = m_shows.begin();
       it != m_shows.end();
       ++it) {
    delete *it;
  }
}

Device*
MediaQueue::findDevice(std::string a_deviceName) const
{
  for (DeviceList::const_iterator deviceIt = m_devices.begin();
       deviceIt != m_devices.end();
       ++deviceIt) {
    if ((*deviceIt)->getName() == a_deviceName) {
      return *deviceIt;
    }
  }
  return 0;
}

void
MediaQueue::schedule(time_t a_begin, time_t a_end,
                     TimeShowMap& a_show, TimeShowMap& a_skip)
{
  ::schedule(m_shows, a_begin, a_end, a_show, a_skip);
}

void
schedule(ShowList& a_shows,
         time_t a_begin, time_t a_end,
         TimeShowMap& a_show, TimeShowMap& a_skip)
{
  TimeShowMMap begins;

  a_show.clear();
  a_skip.clear();

  // for all shows, insert all apparances of the show to the begin/end-maps
  for (ShowList::iterator it = a_shows.begin();
       it != a_shows.end();
       ++it) {
    time_t timeNow = a_begin;
    time_t time;

    int round = 0;
    
    do {
      time = (round == 0) ? (*it)->lastBegun(timeNow) : (*it)->nextBegins(timeNow);
      if (time) {
        begins.insert(std::make_pair(time, *it));
      }
      if (timeNow == a_begin) {
        time_t lastTime = (*it)->lastEnded(timeNow);
        if (lastTime >= timeNow && lastTime) {
          time_t showTime = (*it)->lastBegun(timeNow);
          begins.insert(std::make_pair(showTime, *it));
        }
      }
      timeNow = time + 1;
      ++round;
    } while (round < 2 || (time && timeNow < a_end));
  }

  if (begins.size()) {
    typedef std::pair<time_t, Show*> ShowOccurrance;
    typedef std::map<ShowOccurrance, bool> ShownMap;
    ShownMap shown;

    // we'll put every show to be a candidate for showing
    for (TimeShowMMap::iterator it1 = begins.begin();
         it1 != begins.end();
         ++it1) {
      shown[*it1] = true;
    }
    
    // first remove conflicting shows
    for (TimeShowMMap::iterator it1 = begins.begin();
         it1 != begins.end();
         ++it1) {
      for (TimeShowMMap::iterator it2 = begins.begin();
           it2 != begins.end();
           ++it2) {
        if (it1->second != it2->second) {
          if (it1->second->getDevice() == it2->second->getDevice()) {
            if (overlaps(it1->first, it1->second->getMinutes(),
                         it2->first, it2->second->getMinutes())) {
              if (it1->second->getPriority() < it2->second->getPriority()) {
                shown[*it2] = false;
              } else {
                shown[*it1] = false;
              }
            }
          }
        }
      }
    }

    // now check if some shows don't conflict with other shows anymore
    for (TimeShowMMap::iterator it1 = begins.begin();
         it1 != begins.end();
         ++it1) {
      bool show = shown[*it1];
      if (!show) {
        show = true;
        for (TimeShowMMap::iterator it2 = begins.begin();
             it2 != begins.end() && show;
             ++it2) {
          if (it1->second != it2->second) {
            if (it1->second->getDevice() == it2->second->getDevice()) {
              if (overlaps(it1->first, it1->second->getMinutes(),
                           it2->first, it2->second->getMinutes()) &&
                  shown[*it2]) {
                if (it1->second->getPriority() > it2->second->getPriority()) {
                  show = false;
                }
              }
            }
          }
        }
      }
      shown[*it1] = show;
    }

    for (TimeShowMMap::iterator it = begins.begin();
         it != begins.end();
         ++it) {
      if (it->first + it->second->getMinutes() * 60 >= a_begin &&
          (it->first + it->second->getMinutes() * 60 < a_end ||
           it->first < a_end)) {
        // append to apropriate list
        if (shown[*it]) {
          a_show[it->first] = it->second;
        } else {
          a_skip[it->first] = it->second;
        }
      }
    }
  }
}

void
MediaQueue::addShow(Show* a_show)
{
  m_shows.push_back(a_show);
}

MediaQueue::iterator
MediaQueue::begin()
{
  return m_shows.begin();
}

MediaQueue::iterator
MediaQueue::end()
{
  return m_shows.end();
}

MediaQueue::const_iterator
MediaQueue::begin() const
{
  return m_shows.begin();
}

MediaQueue::const_iterator
MediaQueue::end() const
{
  return m_shows.end();
}

std::ostream&
operator<<(std::ostream& a_stream,
           const MediaQueue& a_queue)
{
  for (MediaQueue::const_iterator it = a_queue.begin();
       it != a_queue.end();
       ++it) {
    a_stream << **it << " " << *(*it)->getOccurrence() << std::endl;
  }
  return a_stream;
}

void
showSchedule(std::ostream& a_stream, const TimeShowMap& a_times)
{
  int total = 0;
  for (TimeShowMap::const_iterator it = a_times.begin();
       it != a_times.end();
       ++it) {
    a_stream << formatTime(it->first, "%a %Y-%m-%d %H:%M") << " | " << *it->second << std::endl;
    total += it->second->getMinutes();
  }
  a_stream << "total " << total / 60 << "h" << total % 60 << "m\n";
}



// changelog:
//
// $Log: mediaqueue.cc,v $
// Revision 1.18  2002/05/27 20:53:45  flux
// Now shows totals in showSchedule
//
// Revision 1.17  2002/05/27 18:38:48  flux
// Removed ends, it was not used for anything.
// Fixed algorithm error. Different show times are now handled as separate.
//
// Revision 1.16  2002/05/26 20:50:53  flux
// Added a clipping-stage to schedule.
// Start scheduling 24 hours before requested, to get better view of overlapping shows.
//
// Revision 1.15  2002/05/20 19:25:34  flux
// Better (??) iterator comparison.
// Removed code out-ifdeffing
//
// Revision 1.14  2002/05/20 19:10:12  flux
// Algorithm optimization turned out to be bogus - fixed
//
// Revision 1.13  2002/05/19 18:36:29  flux
// Renamed a variable to reflect the usage better
//
// Revision 1.12  2002/05/17 20:49:52  flux
// Fixed the schedule-algoritm
//
// Revision 1.11  2002/05/17 20:30:01  flux
// Minor optimization in schedule calculation
//
// Revision 1.10  2002/05/17 18:58:33  flux
// Updated copyright message
//
// Revision 1.9  2002/05/15 17:46:21  flux
// Changed the formatting of dates a little
//
// Revision 1.8  2001/11/15 15:31:56  flux
// Add stream-argument to showSchedule
//
// Revision 1.7  2001/11/15 15:10:35  flux
// Moved loading configuration out of MediaQueue
//
// Revision 1.6  2001/11/15 14:54:08  flux
// Added Device::findDevice
//
// Revision 1.5  2001/11/15 14:34:22  flux
// Added MediaQueue::transferQueue to transfer existing devices from old MediaQueue instances.
// Added an extra argument MediaQueue* old to MediaQueue::load for the same reason.
//
// Revision 1.4  2001/11/15 08:56:01  flux
// Added a hidden low-order priority consisting of the line number in the configuration file
//
// Revision 1.3  2001/11/15 08:53:00  flux
// Added support for one-shot recordings
//
// Revision 1.2  2001/11/13 19:54:30  flux
// Support for 7 = 0
//
// Revision 1.1  2001/11/04 13:30:58  flux
// First version
//
