// Copyright Erkki Seppl 2002
// $Id: socket.cc,v 1.8 2003/02/16 01:38:46 flux Exp $
// $Author: flux $ 
// $Revision: 1.8 $

static char* gs_id __attribute__ ((unused)) ="$Id: socket.cc,v 1.8 2003/02/16 01:38:46 flux Exp $"; /// ID for CVS/RCS

#include <network/socket.hh>
#include <network/error.hh>
#include <network/address.hh>
#include <network/stream.hh>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

namespace Network {
  Socket::Socket() :
    Selectable(),
    m_fd(-1)
  {
    // nothing
  }

  Socket::Socket(int a_fd) :
    Selectable(),
    m_fd(a_fd)
  {
    // nothing
  }

  Socket::Socket(const Socket& a_other) : Selectable()
  {
    m_fd = a_other.copyFd();
  }

  Socket&
  Socket::operator=(const Socket& a_other)
  {
    if (this != &a_other) {
      m_stream.reset(0);
      if (m_fd != -1) {
        close(m_fd);
      }
      m_fd = a_other.copyFd();
    }
    return *this;
  }
  
  int
  Socket::copyFd() const
  {
    // !!!! add locking
    if (m_fd != -1) {
      return dup(m_fd);
    } else {
      return -1;
    }
  }
  
  Socket::~Socket()
  {
    m_stream.reset(0);
    if (m_fd != -1) {
      close(m_fd);
    }
  }

  void
  Socket::drop()
  {
    m_stream.reset(0);
    m_fd = -1;
  }

  int
  Socket::getFd() const
  {
    return m_fd;
  }

  void
  Socket::setFd(int a_fd)
  {
    m_stream.reset(0);
    if (m_fd != -1) {
      close(m_fd);
    }
    m_fd = a_fd;
  }

  bool
  Socket::connect(const Address& a_address, int* a_error, bool a_nonblock)
  {
    int wasMode = fcntl(m_fd, F_GETFL);
    fcntl(m_fd, F_SETFL, (wasMode & ~O_NONBLOCK) | (a_nonblock ? O_NONBLOCK : 0));
    bool rc = ::connect(m_fd, a_address.getAddress(), a_address.getSize()) == 0;
    if (a_error) {
      *a_error = errno;
    }
    fcntl(m_fd, F_SETFL, wasMode);
    return rc;
  }

  Socket
  Socket::accept(int* a_error)
  {
    int fd = ::accept(m_fd, 0, 0);
    if (fd != -1) {
      return Socket(fd);
    } else {
      if (a_error) {
        *a_error = errno;
      }
      return Socket();
    }
  }

  Socket
  Socket::accept(Address& a_address, int* a_error)
  {
    char addr[1024];
    socklen_t size = sizeof(addr);
    int fd = ::accept(m_fd, (sockaddr*) addr, &size);
    if (fd != -1) {
      a_address.set(addr);
      return Socket(fd);
    } else {
      if (a_error) {
        *a_error = errno;
      }
      return Socket();
    }
  }

  Socket::operator bool() const
  {
    return m_fd != -1;
  }

  bool
  Socket::bind(const Address& a_address, int* a_error)
  {
    if (::bind(m_fd, a_address.getAddress(), a_address.getSize()) < 0) {
      if (a_error) {
        *a_error = errno;
      }
      // !!!!
      return false;
    } else {
      return true;
    }
  }

  bool
  Socket::listen(int a_backlog, int* a_error)
  {
    if (::listen(m_fd, a_backlog) < 0) {
      // !!!!
      if (a_error) {
        *a_error = errno;
      }
      return false;
    } else {
      return true;
    }
  }

  int
  Socket::readFull(void* a_buf, int a_count, int* a_error)
  {
    int gotBytes = 0;
    while (a_count > 0) {
      int bytes = ::read(m_fd, (char*) a_buf + gotBytes, a_count);
      if (bytes < 0) {
        if (errno != EINTR && errno != EAGAIN) {
          if (a_error) {
            *a_error = errno;
          }
          return -1;
        }
      } else if (bytes == 0) {
        a_count = 0;
      } else {
        gotBytes += bytes;
        a_count -= bytes;
      }
    }
    return gotBytes;
  }
  
  int
  Socket::writeFull(void* a_buf, int a_count, int* a_error)
  {
    int wroteBytes = 0;
    while (a_count > 0) {
      int bytes = ::write(m_fd, (char*) a_buf + wroteBytes, a_count);
      if (bytes < 0) {
        if (errno != EINTR && errno != EAGAIN) {
          if (a_error) {
            *a_error = errno;
          }
          return -1;
        }
      } else if (bytes == 0) {
        a_count = 0;
      } else {
        wroteBytes += bytes;
        a_count -= bytes;
      }
    }
    return wroteBytes;
  }
  
  int
  Socket::recv(void* a_buf, int a_count, int* a_error)
  {
    int b = ::recv(m_fd, a_buf, a_count, 0);
    if (b < 0 && a_error) {
      *a_error = errno;
    }
    return b;
  }
  
  int
  Socket::recv(void* a_buf, int a_count, Address& a_address, int* a_error)
  {
    char addr[1024];
    socklen_t size = sizeof(addr);
    int bytes = ::recvfrom(m_fd, a_buf, a_count, 0, (sockaddr*) addr, &size);
    if (bytes >= 0) {
      a_address.set(addr);
    } else {
      if (a_error) {
        *a_error = errno;
      }
    }
    return bytes;
  }

  int
  Socket::send(const void* a_buf, int a_count, int* a_error)
  {
    int b = ::send(m_fd, a_buf, a_count, 0);
    if (b < 0 && a_error) {
      *a_error = errno;
    }
    return b;
  }
  
  int
  Socket::send(const void* a_buf, int a_count, const Address& a_to, int* a_error)
  {
    int b = ::sendto(m_fd, a_buf, a_count, 0, a_to.getAddress(), a_to.getSize());
    if (b < 0 && a_error) {
      *a_error = errno;
    }
    return b;
  }

  Stream&
  Socket::stream()
  {
    if (!m_stream.get()) {
      m_stream.reset(new Stream(m_fd));
    }
    return *m_stream;
  }

  int
  Socket::markSelect(fd_set* a_readFds,
                     fd_set* a_writeFds,
                     fd_set* a_exceptFds)
  {
    if (a_readFds) {
      FD_SET(m_fd, a_readFds);
    }
    if (a_writeFds) {
      FD_SET(m_fd, a_writeFds);
    }
    if (a_exceptFds) {
      FD_SET(m_fd, a_exceptFds);
    }
    return m_fd + 1;
  }
 
  bool
  Socket::checkSelect(const fd_set* a_readFds,
                      const fd_set* a_writeFds,
                      const fd_set* a_exceptFds)
  {
    return (
            (a_readFds && FD_ISSET(m_fd, a_readFds)) ||
            (a_writeFds && FD_ISSET(m_fd, a_writeFds)) ||
            (a_exceptFds && FD_ISSET(m_fd, a_exceptFds))
            );
  }

  void
  Socket::reuseAddress(bool a_reuse)
  {
    int reuse = a_reuse;
    setsockopt(m_fd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse));
  }

}


// changelog:
//
// $Log: socket.cc,v $
// Revision 1.8  2003/02/16 01:38:46  flux
// Added non-blocking connect
//
// Revision 1.7  2002/05/28 18:30:01  flux
// Added support for SO_REUSEADDR
//
// Revision 1.6  2002/05/28 15:53:33  flux
// Added selecting support
//
// Revision 1.5  2002/05/28 15:17:54  flux
// Added writeFull
//
// Revision 1.4  2002/05/28 15:02:59  flux
// Added readFull
//
// Revision 1.3  2002/05/28 14:55:04  flux
// Updated copyright message
//
// Revision 1.2  2002/05/28 07:46:06  flux
// Added connect
//
// Revision 1.1  2002/05/22 08:33:29  flux
// Added some Socket-wrapping
//
